from autohack.core.lib.i18n import *
from typing import Callable
import readchar, inspect, pathlib, time, sys, os


class ANSIHelper:
    CLEAR_LINE = "\x1b[2K\r"
    PREV_LINE = "\x1b[1A"
    NEXT_LINE = "\n"

    TEMPLATE = "\x1b[{}m"

    RESET = "0"
    BOLD = "1"
    DIM = "2"
    ITALIC = "3"
    UNDERLINE = "4"
    REVERSED = "7"
    DELETELINE = "9"

    BLACK = "30"
    RED = "31"
    GREEN = "32"
    YELLOW = "33"
    BLUE = "34"
    MAGENTA = "35"
    CYAN = "36"
    WHITE = "37"

    @staticmethod
    def formatCode(effect: list[str]) -> str:
        return f"{ANSIHelper.TEMPLATE.format(';'.join(effect))}"

    @staticmethod
    def colorText(text: str, effect: str | list[str] | None) -> str:
        effect = [effect] if isinstance(effect, str) else (effect or [])
        return f"{ANSIHelper.formatCode(effect)}{text}{ANSIHelper.formatCode([ANSIHelper.RESET])}"


def ensureDirExists(dirPath: pathlib.Path) -> None:
    dirPath.mkdir(parents=True, exist_ok=True)


def mswindows() -> bool:
    # Trick from subprocess
    try:
        import msvcrt
    except ModuleNotFoundError:
        return False
    else:
        return True


def formatTime(t: time.struct_time = time.localtime()) -> str:
    return time.strftime("%Y%m%d%H%M%S", t)


def writeData(filePath: pathlib.Path, data: bytes) -> None:
    ensureDirExists(filePath.parent)
    open(filePath, "wb").write(data)


def readData(filePath: pathlib.Path) -> bytes:
    return open(filePath, "rb").read()


def inputMessage(prompt: str = "", endl: int = 0, clear: bool = False) -> str:
    showCursor()
    write(prompt, endl=endl, clear=clear)
    try:
        return input()
    finally:
        hideCursor()


def write(*messages: str, endl: int = 0, sep: str = "", clear: bool = False, flush: bool = True) -> None:
    if clear:
        sys.stdout.write(ANSIHelper.CLEAR_LINE)
    message = sep.join(messages)
    sys.stdout.write(message)
    sys.stdout.write("\n" * endl)
    if flush:
        sys.stdout.flush()


def getTranslatedMessage(I18n: I18N, message: str, *args, language: str = "") -> str:
    return I18n.translate(message, language).format(*map(str, args))


def writeMessage(
    I18n: I18N,
    message: str,
    *args,
    language: str = "",
    endl: int = 0,
    clear: bool = False,
    highlight: bool = False,
) -> None:
    message = I18n.translate(message, language).format(*map(str, args))
    if highlight:
        message = ANSIHelper.colorText(message, [ANSIHelper.BOLD, ANSIHelper.RED])
    write(message, endl=endl, clear=clear)


def hideCursor() -> None:
    # https://www.cnblogs.com/chargedcreeper/p/-/ANSI
    write("\x1b[?25l")


def showCursor() -> None:
    write("\x1b[?25h")


def exitProgram(exitCode: int = 0, pure: bool = False) -> None:
    if not pure:
        showCursor()
    sys.exit(exitCode)


def getFunctionInfo(func: Callable) -> tuple[list[type], type]:
    sig = inspect.signature(func)
    params = sig.parameters
    paramTypes = [param.annotation for param in params.values()]
    return (paramTypes, sig.return_annotation)


def getFolderSize(folderPath: pathlib.Path) -> int:
    # return shutil.disk_usage(folderPath).used
    totalSize = 0
    with os.scandir(folderPath) as entries:
        for entry in entries:
            if entry.is_file():
                totalSize += entry.stat().st_size
            elif entry.is_dir():
                totalSize += getFolderSize(pathlib.Path(entry.path))
    return totalSize


def selectionMenu(selectionList: list[str]) -> int:
    currentSelection = 0
    write("Use Up/Down arrows to navigate, Enter to select.", endl=1)

    def updateSelection() -> None:
        for i, selectionItem in enumerate(selectionList):
            write(
                ANSIHelper.colorText(
                    f"{">" if i == currentSelection else " "} {selectionItem}", [ANSIHelper.RED, ANSIHelper.BOLD] if i == currentSelection else []
                ),
                endl=1 if i < len(selectionList) - 1 else 0,
                clear=True,
                # highlight=(i == currentSelection),
            )

    while True:
        updateSelection()
        k = readchar.readkey()
        if k == readchar.key.UP or k == "k":
            currentSelection = currentSelection - 1 if currentSelection > 0 else len(selectionList) - 1
        elif k == readchar.key.DOWN or k == "j":
            currentSelection = currentSelection + 1 if currentSelection < len(selectionList) - 1 else 0
        elif k == readchar.key.ENTER:
            for _ in range(len(selectionList) + 1):
                write(ANSIHelper.CLEAR_LINE)
                write(ANSIHelper.PREV_LINE)
            return currentSelection
        elif k == readchar.key.ESC:
            exitProgram(0)
        for _ in range(len(selectionList) - 1):
            write(ANSIHelper.PREV_LINE)



from autohack.core.exception import *
import subprocess, threading, psutil, time


class CodeRunner:
    class Result:
        def __init__(
            self,
            totalTime: float | None,
            timeOut: bool,
            maxMemory: int | None,
            memoryOut: bool,
            returnCode: int | None,
            stdout: bytes | None,
            stderr: bytes | None,
        ) -> None:
            self.totalTime = totalTime
            self.timeOut = timeOut
            self.maxMemory = maxMemory
            self.memoryOut = memoryOut
            self.returnCode = returnCode
            self.stdout = stdout
            self.stderr = stderr

    def __init__(self):
        self.totalTime = None
        self.timeOut = False
        self.maxMemory = None
        self.memoryOut = False

    def memoryMonitor(self, pid: int, timeLimit: float | None, memoryLimit: int | None) -> None:
        try:
            psutilProcess = psutil.Process(pid)
            startTime = psutilProcess.create_time()
            while True:
                # psutilProcess.cpu_times();
                self.totalTime = time.time() - startTime
                if timeLimit is not None and self.totalTime > timeLimit:
                    self.timeOut = True
                    psutilProcess.kill()
                    return
                # 测出来是资源监视器内存中提交那栏 *1024
                self.maxMemory = max(-1 if self.maxMemory is None else self.maxMemory, psutilProcess.memory_info().vms)
                if memoryLimit is not None and self.maxMemory > memoryLimit:
                    self.memoryOut = True
                    psutilProcess.kill()
                    return
        except psutil.NoSuchProcess:
            return

    def run(
        self,
        *popenargs,
        inputContent: bytes | None = None,
        timeLimit: float | None = None,
        memoryLimit: int | None = None,
        **kwargs,
    ) -> Result:
        returnCode = 0
        stdout = None
        stderr = None
        with subprocess.Popen(*popenargs, **kwargs) as process:
            monitor = threading.Thread(target=self.memoryMonitor, args=(process.pid, timeLimit, memoryLimit))
            monitor.start()
            stdout, stderr = process.communicate(inputContent)  # type: ignore
            returnCode = process.poll()
        return self.Result(self.totalTime, self.timeOut, self.maxMemory, self.memoryOut, returnCode, stdout, stderr)  # type: ignore


def compileCode(compileCommand: list) -> None:
    try:
        process = subprocess.Popen(compileCommand, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    except OSError:
        return
    output = process.communicate()[0]
    if process.returncode != 0:
        raise autohackRuntimeError(output, process.returncode)


def generateInput(generateCommand: list) -> bytes:
    try:
        process = subprocess.Popen(generateCommand, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    except OSError:
        return b""
    dataInput = process.communicate()[0]
    if process.returncode != 0:
        raise autohackRuntimeError(dataInput, process.returncode)
    return dataInput


def generateAnswer(generateCommand: list, dataInput: bytes) -> bytes:
    try:
        process = subprocess.Popen(generateCommand, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    except OSError:
        return b""
    dataAnswer = process.communicate(dataInput)[0]
    if process.returncode != 0:
        raise autohackRuntimeError(dataAnswer, process.returncode)
    return dataAnswer


def runSourceCode(runCommand: list, dataInput: bytes, timeLimit: float | None, memoryLimit: int | None) -> CodeRunner.Result:
    try:
        result = CodeRunner().run(
            runCommand,
            inputContent=dataInput,
            timeLimit=timeLimit,
            memoryLimit=memoryLimit,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
    except OSError:
        return CodeRunner.Result(None, False, None, False, 0, b"", b"")
    return result



from typing import Any
import pathlib
import tarfile
import time

from autohack.core.util.util import *

from loguru import logger as baseLogger


class Logger:
    def __init__(
        self,
        logFolder: pathlib.Path,
        debug: bool = False,
    ) -> None:
        self.logFolder = logFolder
        self.debug = debug
        self.logLevel = "DEBUG" if debug else "INFO"

        ensureDirExists(self.logFolder)

        self.logFilePath = self.logFolder / "latest.log"

        if self.logFilePath.exists():
            try:
                mtime = time.localtime(self.logFilePath.stat().st_mtime)
                dataStr = time.strftime("%Y-%m-%d", mtime)
                idx = 1
                while True:
                    archive_name = self.logFolder / f"{dataStr}-{idx}.log.tar.gz"
                    if not archive_name.exists():
                        break
                    idx += 1

                with tarfile.open(archive_name, "w:gz") as tarf:
                    tarf.add(self.logFilePath, arcname=f"{dataStr}-{idx}.log")
                self.logFilePath.unlink()
            except Exception:
                pass

        baseLogger.remove()

        logFormat = "{time:YYYY-MM-DD HH:mm:ss} | {level} | [{extra[module]}] {message}"
        baseLogger.add(str(self.logFilePath), format=logFormat, level=self.logLevel, encoding="utf-8")

        self.logger = baseLogger.bind(module="logger")

        self.logger.info(f'Log file: "{self.logFilePath}"')
        self.logger.info(f"Log level: {self.logLevel}")
        self.logger.info("Logger initialized.")

    def getLogger(self) -> Any:
        return self.logger

    def getLogFilePath(self) -> pathlib.Path:
        return self.logFilePath



from autohack import __VERSION__
from autohack.core.plugin.checker import *
from autohack.core.constant import *
from autohack.core.exception import *
from autohack.core.path import *
from autohack.core.util.util import *
from autohack.core.util.run import *
from autohack.core.lib.config import *
from autohack.core.lib.logger import *
from autohack.core.lib.i18n import *
import traceback, time, os


class AppCentral:
    def __init__(self, clientID: str, logTime: time.struct_time, debug: bool = False) -> None:
        self.clientID = clientID
        self.logTime = logTime
        self.debug = debug

    def run(self) -> None:
        hideCursor()

        ensureDirExists(CHECKER_FOLDER_PATH)
        ensureDirExists(LOG_FOLDER_PATH)

        # TODO: Debug
        loggerObj = Logger(LOG_FOLDER_PATH, self.debug)
        logger = loggerObj.getLogger().bind(module="autohack")

        I18n = I18N(TRANSLATION_FOLDER_PATH, logger)
        _ = I18n.translate

        if not GLOBAL_CONFIG_FILE_PATH.exists():
            logger.info("Global config file not found. Creating new one.")
            write("Welcome to autohack-next!", endl=1)
            write("A global config file will be created.", endl=1)
            write("Please select your preferred language:", endl=1)
            # for i, (langID) in enumerate(LANGUAGE_MAPS):
            #     write(f"  {i+1}: {langID} / {I18N(TRANSLATION_FOLDER_PATH, logger).translate("language-info", langID)}", 1)
            # showCursor()
            # while True:
            #     result = inputMessage("Enter the number of your preferred language: ", 0, True)
            #     if result.isdigit() and 1 <= int(result) <= len(LANGUAGE_MAPS):
            #         selectedLang = LANGUAGE_MAPS[int(result) - 1]
            #         break
            #     write("Invalid input. Please enter a valid number.")
            #     prevLine()
            selectedLangIndex = selectionMenu(
                [f"{langID} / {I18N(TRANSLATION_FOLDER_PATH, logger).translate("language-info", langID)}" for i, langID in enumerate(LANGUAGE_MAPS)]
            )
            selectedLang = LANGUAGE_MAPS[selectedLangIndex]
            globalConfig = Config(GLOBAL_CONFIG_FILE_PATH, DEFAULT_GLOBAL_CONFIG, logger)
            globalConfig.modifyConfigEntry("language", selectedLang)
            I18n.setDefaultLanguage(selectedLang)
            write(ANSIHelper.CLEAR_LINE)
            write(ANSIHelper.PREV_LINE)
            write(ANSIHelper.CLEAR_LINE)
            write(ANSIHelper.PREV_LINE)
            write(ANSIHelper.CLEAR_LINE)
            writeMessage(I18n, "__main__.language-select.result", _("language-info"), endl=1)
            writeMessage(I18n, "__main__.language-select.info", GLOBAL_CONFIG_FILE_PATH, endl=2)

        globalConfig = Config(GLOBAL_CONFIG_FILE_PATH, DEFAULT_GLOBAL_CONFIG, logger)
        I18n.setDefaultLanguage(globalConfig.getConfigEntry("language"))

        config = Config(
            CONFIG_FILE_PATH,
            DEFAULT_CONFIG,
            logger,
            CONFIG_VALIDATION_EXCLUDE,
            getTranslatedMessage(I18n, "__main__.config-created", CONFIG_FILE_PATH),
        )

        logger.info(f'Data folder path: "{DATA_FOLDER_PATH}"')
        logger.info(f"Client ID: {self.clientID}")
        logger.info(f"Initialized. Version: {__VERSION__}")
        writeMessage(I18n, "__main__.start.version", __VERSION__, self.clientID, endl=2)
        writeMessage(I18n, "__main__.start.data", getHackDataStorageFolderPath(self.clientID, self.logTime), endl=1)
        writeMessage(I18n, "__main__.start.log", loggerObj.getLogFilePath(), endl=1)
        writeMessage(I18n, "__main__.start.export", getExportFolderPath(self.logTime, self.clientID), endl=1)
        writeMessage(I18n, "__main__.start.checker", CHECKER_FOLDER_PATH, endl=2)

        waitTimeBeforeStart = globalConfig.getConfigEntry("wait_time_before_start")
        for i in range(waitTimeBeforeStart, 0, -1):
            writeMessage(I18n, "__main__.countdown", i, clear=True)
            time.sleep(1)

        fileList = [
            [config.getConfigEntry("commands.compile.source"), "__main__.compile.filename.source"],
            [config.getConfigEntry("commands.compile.std"), "__main__.compile.filename.std"],
            [config.getConfigEntry("commands.compile.generator"), "__main__.compile.filename.generator"],
        ]
        for file in fileList:
            writeMessage(I18n, "__main__.compile.doing", _(file[1]), clear=True)
            try:
                compileCode(file[0])
            except autohackRuntimeError as e:
                logger.error(
                    f"{_(file[1], LOGGER_LANGUAGE_ID).capitalize()} compilation failed with return code {e.returnCode} and message:\n{e.output.decode(errors="ignore")}"
                )
                writeMessage(I18n, "__main__.compile.error", _(file[1]).capitalize(), e.returnCode, endl=2, clear=True, highlight=True)
                write(e.output.decode(errors="ignore"))
                exitProgram(1)
            else:
                logger.debug(f"{_(file[1], LOGGER_LANGUAGE_ID).capitalize()} compiled successfully.")
        writeMessage(I18n, "__main__.compile.finish", endl=1, clear=True)

        writeMessage(I18n, "__main__.activate-checker.doing", config.getConfigEntry("checker.name"))
        currentChecker: checkerType = lambda l, o, a, ar: (False, _("__main__.activate-checker.no-checker-message"))
        deactivateFunc: deactivateType = emptyDeactivate
        try:
            getCheckerResult = getChecker(CHECKER_FOLDER_PATH, config.getConfigEntry("checker.name"), config.getConfigEntry("checker.args"))
            currentChecker = getCheckerResult[0]
            deactivateFunc = getCheckerResult[1]
        except Exception as e:
            logger.critical(f"{e}")
            writeMessage(I18n, "__main__.activate-checker.failed", endl=1, clear=True, highlight=True)
            traceback.print_exc()
            exitProgram(1)
        writeMessage(I18n, "__main__.activate-checker.finish", config.getConfigEntry("checker.name"), endl=2, clear=True)

        dataCount, errorDataCount = 0, 0
        lastStatusError = False
        generateCommand = config.getConfigEntry("commands.run.generator")
        stdCommand = config.getConfigEntry("commands.run.std")
        sourceCommand = config.getConfigEntry("commands.run.source")
        timeLimit = config.getConfigEntry("time_limit") / 1000
        memoryLimit = config.getConfigEntry("memory_limit") * 1024 * 1024
        inputFilePath = config.getConfigEntry("paths.input")
        answerFilePath = config.getConfigEntry("paths.answer")
        outputFilePath = config.getConfigEntry("paths.output")
        maximumDataLimit = config.getConfigEntry("maximum_number_of_data")
        errorDataLimit = config.getConfigEntry("error_data_number_limit")
        refreshSpeed = globalConfig.getConfigEntry("refresh_speed")
        checkerArgs = config.getConfigEntry("checker.args")

        timeLimit = None if timeLimit == 0 else timeLimit
        memoryLimit = None if memoryLimit == 0 else memoryLimit

        def updateStatus(total: float, averagePerS: float, averagePerData: float, addtional: str) -> None:
            # write(
            #     f"Time taken: {total:.2f} seconds, average {averagePerS:.2f} data per second, {averagePerData:.2f} second per data.{addtional}",
            #     clear=True,
            # )
            writeMessage(I18n, "__main__.status", f"{total:.2f}", f"{averagePerS:.2f}", f"{averagePerData:.2f}", clear=True)
            write(addtional)

        write(endl=1)
        updateStatus(0.0, 0.0, 0.0, " (0%)" if maximumDataLimit > 0 else "")
        write(ANSIHelper.PREV_LINE)

        startTime = time.time()

        while (maximumDataLimit <= 0 or dataCount < maximumDataLimit) and (errorDataLimit <= 0 or errorDataCount < errorDataLimit):
            dataInput = b""
            dataAnswer = b""

            dataCount += 1

            try:
                # write(f"{dataCount}: Generate input.", clear=True)
                writeMessage(I18n, "__main__.main.generate-input", dataCount, clear=True)
                logger.debug(f"Generating data {dataCount}.")
                dataInput = generateInput(generateCommand)
            except autohackRuntimeError as e:
                logger.error(f"Input generation failed with return code {e.returnCode}.")
                writeMessage(I18n, "__main__.main.generate-input-failed", e.returnCode, endl=1, clear=True, highlight=True)
                inputExportPath = getExportDataPath(getExportFolderPath(self.logTime, self.clientID), "input")
                writeData(inputExportPath, e.output)
                writeMessage(I18n, "__main__.main.save-input-data", inputExportPath, clear=True)
                exitProgram(1)

            try:
                # write(f"{dataCount}: Generate answer.", clear=True)
                writeMessage(I18n, "__main__.main.generate-answer", dataCount, clear=True)
                logger.debug(f"Generating answer for data {dataCount}.")
                dataAnswer = generateAnswer(stdCommand, dataInput)
            except autohackRuntimeError as e:
                logger.error(f"Answer generation failed with return code {e.returnCode}.")
                writeMessage(I18n, "__main__.main.generate-answer-failed", e.returnCode, endl=1, clear=True, highlight=True)
                inputExportPath = getExportDataPath(getExportFolderPath(self.logTime, self.clientID), "input")
                writeData(inputExportPath, dataInput)
                writeMessage(I18n, "__main__.main.save-input-data", inputExportPath, endl=1, clear=True)
                answerExportPath = getExportDataPath(getExportFolderPath(self.logTime, self.clientID), "answer")
                writeData(answerExportPath, e.output)
                writeMessage(I18n, "__main__.main.save-answer-data", answerExportPath, clear=True)
                exitProgram(1)

            # write(f"{dataCount}: Run source code.", clear=True)
            writeMessage(I18n, "__main__.main.run-source", dataCount, clear=True)
            logger.debug(f"Run source code for data {dataCount}.")
            result = runSourceCode(sourceCommand, dataInput, timeLimit, memoryLimit)
            if result.stdout is None:
                result.stdout = b""
            if result.stderr is None:
                result.stderr = b""

            # TODO: Refresh when running exe. Use threading or async?
            if dataCount % refreshSpeed == 0 or lastStatusError:
                lastStatusError = False
                currentTime = time.time()
                write(endl=1)
                # write(
                #     f"Time taken: {currentTime - startTime:.2f} seconds, average {dataCount/(currentTime - startTime):.2f} data per second, {(currentTime - startTime)/dataCount:.2f} second per data.{f" ({dataCount*100/maximumDataLimit:.0f}%)" if maximumDataLimit > 0 else ""}",
                #     clear=True,
                # )
                updateStatus(
                    currentTime - startTime,
                    dataCount / (currentTime - startTime),
                    (currentTime - startTime) / dataCount,
                    f" ({dataCount*100/maximumDataLimit:.0f}%)" if maximumDataLimit > 0 else "",
                )
                write(ANSIHelper.PREV_LINE)

            saveData, termMessage, logMessage, extMessage, exitAfterSave = (False, "", "", None, False)

            if result.memoryOut:
                saveData = True
                logMessage = f"Memory limit exceeded for data {dataCount}."
                termMessage = getTranslatedMessage(I18n, "__main__.main.memory-limit-exceeded", dataCount)
                if result.maxMemory is not None:
                    extMessage = getTranslatedMessage(I18n, "__main__.main.memory-limit-exceeded-extra", f"{result.maxMemory / 1024 / 1024:.4f}")
            elif result.timeOut:
                saveData = True
                logMessage = f"Time limit exceeded for data {dataCount}."
                termMessage = getTranslatedMessage(I18n, "__main__.main.time-limit-exceeded", dataCount)
                if result.totalTime is not None:
                    extMessage = getTranslatedMessage(I18n, "__main__.main.time-limit-exceeded-extra", f"{result.totalTime*1000:.4f}")
            elif result.returnCode != 0:
                saveData = True
                logMessage = f"Runtime error for data {dataCount} with return code {result.returnCode}."
                termMessage = getTranslatedMessage(I18n, "__main__.main.runtime-error", dataCount)
                if result.returnCode is not None:
                    extMessage = getTranslatedMessage(I18n, "__main__.main.runtime-error-extra", f"{result.returnCode}")

            checkerResult = (False, _("__main__.main.checker-not-executed"))
            try:
                checkerResult = currentChecker(dataInput, result.stdout, dataAnswer, checkerArgs)
            except Exception as e:
                saveData = True
                termMessage = getTranslatedMessage(I18n, "__main__.main.checker-error-without-exception", dataCount)
                logMessage = f"Checker error for data {dataCount}. Exception: {e}"
                extMessage = f"{_("__main__.main.checker-error-extra-message")}\n{traceback.format_exc()}"
                checkerResult = (False, _("__main__.main.checker-exception-occurred"))
                exitAfterSave = True

            if not saveData and not checkerResult[0]:
                saveData = True
                termMessage = getTranslatedMessage(I18n, "__main__.main.wrong-answer", dataCount)
                logMessage = f"Wrong answer for data {dataCount}. Checker output: {checkerResult[1]}"
                extMessage = checkerResult[1]

            if saveData:
                lastStatusError = True
                errorDataCount += 1
                writeData(getHackDataFilePath(getHackDataStorageFolderPath(self.clientID, self.logTime), errorDataCount, inputFilePath), dataInput)
                writeData(getHackDataFilePath(getHackDataStorageFolderPath(self.clientID, self.logTime), errorDataCount, answerFilePath), dataAnswer)
                writeData(
                    getHackDataFilePath(getHackDataStorageFolderPath(self.clientID, self.logTime), errorDataCount, outputFilePath), result.stdout
                )
                write(f"[{errorDataCount}]: {termMessage}", endl=1, clear=True)
                if extMessage is not None and extMessage != "":
                    write(f"{(len(f'[{errorDataCount}]: ')-3)*' '} - {extMessage}", endl=1, clear=True)
                logger.info(f"{logMessage}")

            if exitAfterSave:
                writeMessage(I18n, "__main__.main.checker-failed-exit", clear=True, highlight=True)
                exitProgram(0)

        endTime = time.time()

        writeMessage(I18n, "__main__.main.finish", dataCount, errorDataCount, endl=1, clear=True)
        # write(
        #     f"Time taken: {endTime - startTime:.2f} seconds, average {dataCount/(endTime - startTime):.2f} data per second, {(endTime - startTime)/dataCount:.2f} second per data.",
        #     2,
        #     True,
        # )
        updateStatus(endTime - startTime, dataCount / (endTime - startTime), (endTime - startTime) / dataCount, "")
        write(endl=2)

        # if errorDataCount == 0:
        #     shutil.rmtree(getHackDataStorageFolderPath(clientID))
        #     write("No error data found. Hack data folder removed.", 1)
        #     logger.info("No error data found. Hack data folder removed.")

        dataFolderMaxSize = globalConfig.getConfigEntry("data_folder_max_size")
        # print(getFolderSize(HACK_DATA_STORAGE_FOLDER_PATH) / 1024 / 1024, " ", dataFolderMaxSize)
        if HACK_DATA_STORAGE_FOLDER_PATH.exists() and getFolderSize(HACK_DATA_STORAGE_FOLDER_PATH) > dataFolderMaxSize * 1024 * 1024:
            logger.warning(f"Hack data storage folder size exceeds {dataFolderMaxSize} MB: {HACK_DATA_STORAGE_FOLDER_PATH}")
            # write(f"Warning: Hack data storage folder size exceeds {DATA_FOLDER_MAX_SIZE} MB: {HACK_DATA_STORAGE_FOLDER_PATH}", 2)
            writeMessage(I18n, "__main__.data-folder-size-warning", dataFolderMaxSize, HACK_DATA_STORAGE_FOLDER_PATH, endl=2, highlight=True)

        writeMessage(I18n, "__main__.deactivate-checker.doing")
        try:
            deactivateFunc(config.getConfigEntry("checker.args"))
        except Exception as e:
            logger.error(f"Checker deactivation failed with exception: {e}")
            writeMessage(I18n, "__main__.deactivate-checker.failed", endl=1, clear=True, highlight=True)
            traceback.print_exc()
            exitProgram(1)
        writeMessage(I18n, "__main__.deactivate-checker.finish", endl=1, clear=True)

        writeMessage(I18n, "__main__.post-command", endl=1)
        os.system(config.getConfigEntry("command_at_end"))
        logger.info("Finished.")



import platformdirs, pathlib, time, os

dirs = platformdirs.PlatformDirs("autohack", "Gavin", version="v1")

DATA_FOLDER_PATH = pathlib.Path(os.getcwd()) / ".autohack"

HACK_DATA_STORAGE_FOLDER_PATH = DATA_FOLDER_PATH / "datastorage"

CHECKER_FOLDER_PATH = DATA_FOLDER_PATH / "checkers"

EXPORT_FOLDER_PATH = DATA_FOLDER_PATH / "export"

LOG_FOLDER_PATH = DATA_FOLDER_PATH / "logs"

CONFIG_FILE_PATH = DATA_FOLDER_PATH / "config.json"

GLOBAL_DATA_FOLDER_PATH = pathlib.Path(dirs.user_data_dir)

GLOBAL_CONFIG_FILE_PATH = GLOBAL_DATA_FOLDER_PATH / "config.json"

TRANSLATION_FOLDER_PATH = pathlib.Path(__file__).parent.parent / "assets" / "i18n"


def getHackDataStorageFolderPath(clientID: str, startTime: time.struct_time) -> pathlib.Path:
    dateStr = time.strftime("%Y-%m-%d", startTime)
    timeStr = time.strftime("%H-%M-%S", startTime)
    return HACK_DATA_STORAGE_FOLDER_PATH / dateStr / f"{timeStr}_{clientID}"


def getHackDataFilePath(hackDataStorageFolder: pathlib.Path, dataID: int, filePath: str) -> pathlib.Path:
    return hackDataStorageFolder / filePath.replace("$(id)", str(dataID))


def getExportFolderPath(startTime: time.struct_time, clientID: str | None = None) -> pathlib.Path:
    dateStr = time.strftime("%Y-%m-%d", startTime)
    timeStr = time.strftime("%H-%M-%S", startTime)
    if clientID is None:
        return EXPORT_FOLDER_PATH / dateStr / timeStr
    return EXPORT_FOLDER_PATH / dateStr / f"{timeStr}_{clientID}"


def getExportDataPath(exportFolder: pathlib.Path, filePath: str) -> pathlib.Path:
    return exportFolder / filePath



from typing import Any

DEFAULT_CONFIG: dict[str, Any] = {
    "maximum_number_of_data": 0,
    # ms
    "time_limit": 1000,
    # MiB
    "memory_limit": 256,
    "error_data_number_limit": 1,
    "paths": {
        "input": "$(id)/input",
        "answer": "$(id)/answer",
        "output": "$(id)/output",
    },
    "commands": {
        "compile": {
            "source": [
                "g++",
                "source.cpp",
                "-o",
                "source",
                "-O2",
            ],
            "std": [
                "g++",
                "std.cpp",
                "-o",
                "std",
                "-O2",
            ],
            "generator": [
                "g++",
                "generator.cpp",
                "-o",
                "generator",
                "-O2",
            ],
        },
        "run": {
            "source": [
                "./source",
            ],
            "std": [
                "./std",
            ],
            "generator": [
                "./generator",
            ],
        },
    },
    "checker": {
        "name": "builtin_basic",
        "args": {},
    },
    "command_at_end": "",
}


# Do not modify checker args here, as they may contain user data.
# They are excluded from config validation for the same reason.
CONFIG_VALIDATION_EXCLUDE = ["checker.args"]

DEFAULT_GLOBAL_CONFIG = {
    "language": "en_US",
    "refresh_speed": 10,
    "wait_time_before_start": 3,
    "data_folder_max_size": 256,  # MB
}

# empty: For developers to test missing translations.
LANGUAGE_MAPS = ["en_US", "zh_CN", "empty"]

LOGGER_LANGUAGE_ID = "en_US"



from autohack.core.util.util import *
from typing import Any
import pathlib, json5, os


class Config:
    def __init__(
        self,
        configFilePath: pathlib.Path,
        defaultConfig: dict[str, Any],
        logger: Any,
        configValidationExclude: list[str] = [],
        messageOnCreate: str | None = None,
    ) -> None:
        self.defaultConfig = defaultConfig
        self.configValidationExclude = configValidationExclude
        self.configFilePath = configFilePath
        self.logger = logger.bind(module="config")
        self.logger.info(f'Config file path: "{self.configFilePath}"')
        configFileExists = self.configFileExists()
        self.config = self.loadConfig()
        if not configFileExists and messageOnCreate is not None:
            write(messageOnCreate)
            exitProgram(0)

    def configFileExists(self) -> bool:
        return self.configFilePath.exists()

    def loadConfig(self) -> dict[str, Any]:
        if not os.path.exists(self.configFilePath):
            ensureDirExists(self.configFilePath.parent)
            json5.dump(self.defaultConfig, open(self.configFilePath, "w", encoding="utf-8"), indent=4, quote_keys=True, trailing_commas=False)
            self.logger.info("Config file created.")

        config = json5.load(open(self.configFilePath, "r", encoding="utf-8"))

        # if self.defaultConfig["_version"] > config.get("_version", 0):
        #     mergedConfig = self.mergeConfigs(config, self.defaultConfig)
        #     mergedConfig["_version"] = self.defaultConfig["_version"]
        #     json5.dump(mergedConfig, open(self.configFilePath, "w", encoding="utf-8"), indent=4, quote_keys=True, trailing_commas=False)
        #     write(f"Config file {self.configFilePath} updated to version {self.defaultConfig['_version']}.", 2)
        #     self.logger.info("Config file updated.")
        #     config = mergedConfig

        mergedConfig = self.mergeConfigs(config, self.defaultConfig, self.configValidationExclude, "")
        json5.dump(mergedConfig, open(self.configFilePath, "w", encoding="utf-8"), indent=4, quote_keys=True, trailing_commas=False)
        config = mergedConfig

        self.logger.info("Config file loaded.")
        return config

    def mergeConfigs(self, old: dict[str, Any], newDefault: dict[str, Any], configValidationExclude: list[str], keyName: str) -> dict[str, Any]:
        merged = {}

        for key in newDefault:
            if key in old:
                if isinstance(newDefault[key], dict) and isinstance(old[key], dict):
                    newKeyName = f"{keyName}.{key}" if keyName else key
                    if newKeyName not in configValidationExclude:
                        merged[key] = self.mergeConfigs(old[key], newDefault[key], configValidationExclude, newKeyName)
                    else:
                        merged[key] = old[key]
                else:
                    if type(old[key]) is type(newDefault[key]):
                        merged[key] = old[key]
                    else:
                        merged[key] = newDefault[key]
            else:
                merged[key] = newDefault[key]

        return merged

    def getConfigEntry(self, entryName: str) -> Any:
        entryTree = entryName.split(".")
        result = self.config

        for entryItem in entryTree:
            result = result.get(entryItem, None)
            if result is None:
                break

        self.logger.debug(f'Get config entry: "{entryName}" = "{result}"')
        return result

    def modifyConfigEntry(self, entryName: str, newValue: Any) -> bool:
        """Returns True if the entry was modified, False if it does not exist."""
        entryTree = entryName.split(".")
        currentLevel = self.config

        for level in entryTree[:-1]:
            if not isinstance(currentLevel, dict) or level not in currentLevel:
                return False
            currentLevel = currentLevel[level]
        lastLevel = entryTree[-1]
        if not isinstance(currentLevel, dict) or lastLevel not in currentLevel:
            return False
        currentLevel[lastLevel] = newValue

        json5.dump(self.config, open(self.configFilePath, "w", encoding="utf-8"), indent=4, quote_keys=True, trailing_commas=False)
        self.logger.debug(f'Modify entry: "{entryName}" = "{newValue}"')
        return True



__VERSION__ = None
try:
    from importlib.metadata import version

    __VERSION__ = version("autohack-next")
except Exception:
    pass


def entrypoint():
    from argparse import ArgumentParser
    import secrets, time
    import colorama
    from autohack.core.central import AppCentral

    colorama.just_fix_windows_console()

    argsParser = ArgumentParser(
        prog="autohack",
        description="autohack-next - Automated hack data generator",
    )
    argsParser.add_argument("--version", "-v", action="version", version="Unknown" if __VERSION__ is None else __VERSION__)
    argsParser.add_argument("--debug", "-d", action="store_true", help="Enable debug mode", default=False)
    argsParser.add_argument("--reset", action="store_true", help="Reset all configuration", default=False)

    args = argsParser.parse_args()

    if args.reset:
        from autohack.core.path import GLOBAL_CONFIG_FILE_PATH
        import sys

        if GLOBAL_CONFIG_FILE_PATH.exists():
            GLOBAL_CONFIG_FILE_PATH.unlink()
        sys.stdout.write("Configuration reset successfully.")
        return

    # 使用 8位 短随机字符代替过长的 UUID，防止冗长同时保证多个实例独立
    app = AppCentral(secrets.token_hex(4), time.localtime(), debug=args.debug)

    try:
        app.run()
    except KeyboardInterrupt:
        import sys

        sys.stdout.write("\nKeyboard interrupt received. Exiting...\n")

    # TODO: Keyboard interrupt handling



from autohack.core.constant import *


class autohackRuntimeError(Exception):
    def __init__(self, output: bytes, returnCode: int) -> None:
        self.output = output
        self.returnCode = returnCode



from typing import Any
from autohack.core.constant import *
import logging, pathlib, json5, os


class I18N:
    def __init__(
        self,
        translationFileDir: pathlib.Path,
        logger: Any,
    ) -> None:
        self.translationFileDir = translationFileDir
        self.logger = logger.bind(module="i18n")
        self.translations = {}
        self.setDefaultLanguage(LOGGER_LANGUAGE_ID)

    def setDefaultLanguage(self, language: str) -> None:
        self.logger.info(f'Default language: "{language}"')
        self.defaultLanguage = language
        if language not in self.translations:
            self.loadTranslation(language)

    def loadTranslation(self, language: str) -> None:
        self.translationFile = self.translationFileDir / f"{language}.json"
        self.logger.info(f'Translation file: "{self.translationFile}"')
        self.translations[language] = self.loadTranslationFile(self.translationFile)

    def loadTranslationFile(self, filePath: pathlib.Path) -> dict[str, str]:
        if not os.path.exists(filePath):
            self.logger.critical("Translation file not found.")
            raise FileNotFoundError(f"Translation file {filePath} not found.")
        translations = json5.load(open(filePath, "r", encoding="utf-8"))

        self.logger.info("Translation file loaded.")
        return translations

    def translate(self, key: str, language: str = "") -> str:
        if language == "":
            language = self.defaultLanguage
        if language not in self.translations:
            self.loadTranslation(language)
        return self.translations[language].get(key, key)
